Package: matchpointR
Type: Package
Title: Tidy Access to Women's Tennis Association (WTA) Data
Version: 0.1.0
Authors@R: person(
    given = "Alejandro",
    family = "Navas González",
    email = "angnar@telaris.es",
    role = c("aut", "cre"),
    comment = c(alias = "Angnar"))
Description: Scrapes and tidies publicly available data from the
    Women's Tennis Association website (<https://www.wtatennis.com>).
    Provides helpers to retrieve player biographies, singles and doubles
    career overviews, match histories, live rankings and aggregate
    statistics. Dynamic pages are rendered through a headless 'Chrome'
    session so 'JavaScript'-generated content is fully captured, and
    all outputs are returned as tidy data frames suitable for downstream
    analysis or visualisation.
License: Apache License (>= 2)
URL: https://github.com/Angnar-97/matchpointR
BugReports: https://github.com/Angnar-97/matchpointR/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: chromote, cli, jsonlite, magick, purrr, rvest, stringr,
        tibble, xml2
Suggests: httr2, knitr, rmarkdown, rsvg, testthat (>= 3.0.0), withr
Config/testthat/edition: 3
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-26 14:34:10 UTC; User
Author: Alejandro Navas González [aut, cre] (alias: Angnar)
Maintainer: Alejandro Navas González <angnar@telaris.es>
Repository: CRAN
Date/Publication: 2026-04-28 19:40:12 UTC
Built: R 4.6.0; ; 2026-04-28 20:31:14 UTC; unix

# Generated by roxygen2: do not edit by hand

export(wta_get_player_basics)
export(wta_get_player_matches)
export(wta_get_player_overview)
export(wta_get_rankings)
export(wta_player_url)
importFrom(chromote,ChromoteSession)
importFrom(purrr,list_rbind)
importFrom(purrr,map)
importFrom(rvest,html_attr)
importFrom(rvest,html_node)
importFrom(rvest,html_nodes)
importFrom(rvest,html_text)
importFrom(stringr,str_extract)
importFrom(stringr,str_squish)
importFrom(tibble,tibble)
importFrom(xml2,read_html)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# matchpointR <a href="https://github.com/Angnar-97/matchpointR"><img src="man/figures/logo.png" align="right" height="120" alt="matchpointR logo"/></a>

<!-- badges: start -->
[![R-CMD-check](https://github.com/Angnar-97/matchpointR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/Angnar-97/matchpointR/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/matchpointR)](https://CRAN.R-project.org/package=matchpointR)
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

**matchpointR** turns the public pages of the Women's Tennis Association
(<https://www.wtatennis.com>) into tidy data frames. It ships helpers for
player biographies, career highlights, full match histories and live
rankings.

Dynamic content (matches, rankings) is rendered through a headless
Chrome session via [`chromote`](https://cran.r-project.org/package=chromote),
so JavaScript-generated sections are fully captured before parsing. Where
the WTA site exposes structured `schema.org` JSON-LD data, `matchpointR`
reads from that in preference to CSS selectors for resilience against
site redesigns.

## Installation

You can install the development version of `matchpointR` from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("Angnar-97/matchpointR")
```

Or, once it lands on CRAN:

``` r
install.packages("matchpointR")
```

You will also need a working Chrome/Chromium install for the dynamic
scrapers, which is managed automatically by `chromote`.

## Usage

``` r
library(matchpointR)

# Build a canonical player URL from id + slug
url <- wta_player_url(320301, "katerina-siniakova")

# Player biography (name, nationality, headshot, flag, ...)
wta_get_player_basics(url, download_images = FALSE)

# Career highlights (ranks, titles, prize money)
wta_get_player_overview(url)

# Full match history (walks the "Show more" button)
matches_url <- wta_player_url(320301, "katerina-siniakova", "matches")
wta_get_player_matches(matches_url)

# Live rankings
wta_get_rankings("singles", top = 50)
```

## Terms of service and fair use

`matchpointR` reads **publicly accessible** pages of
<https://www.wtatennis.com> using a single headless browser session per
call. Before using this package — especially at scale — you are
responsible for checking and complying with the WTA website's
[Terms and Conditions](https://www.wtatennis.com/terms-and-conditions).
The package does **not** bypass paywalls, authentication walls or any
technical access controls, does not interact with user accounts, and
exposes no bulk-download helpers. Please scrape considerately: the
`chromote` session hits the site as a real browser, so iterating over
many players or rankings days back-to-back is equivalent to a human
browsing the site rapidly. Add explicit `Sys.sleep()` between calls in
loops.

## Scope and caveats

- HTML selectors may drift as the site is redesigned. Where possible
  `matchpointR` reads from the page's `schema.org` JSON-LD block for
  stability. File an issue on GitHub if a function stops returning data.
- Functions return everything as character to stay faithful to the
  rendered page; cast to numeric or date in a follow-up step.
- Tour-wide statistics leaderboards are not yet implemented; tracked in
  [issue #1](https://github.com/Angnar-97/matchpointR/issues/1).

## License

[Apache License (>= 2)](https://www.apache.org/licenses/LICENSE-2.0).

## Author

Alejandro Navas González (Angnar).

#' Get the current WTA rankings
#'
#' Scrapes the rankings table at
#' <https://www.wtatennis.com/rankings/singles> (or `/doubles`) and returns
#' a tidy tibble. The initial page renders the first 50 rows; increase the
#' browser dwell time with `wait` if the widget hasn't hydrated yet.
#'
#' @param type Character. One of `"singles"`, `"doubles"`. Defaults to
#'   `"singles"`.
#' @param top Integer. Limit the output to the top `N` ranked players.
#'   `NULL` (default) keeps every row rendered by the page.
#' @param wait Numeric. Seconds to wait for the rankings widget to hydrate
#'   after navigation. Defaults to 12.
#'
#' @return A [tibble::tibble()] with one row per player and columns:
#'   `rank`, `player_id`, `player`, `country`, `age`,
#'   `tournaments_played`, `points`.
#' @export
#' @examplesIf interactive()
#' wta_get_rankings("singles", top = 50)
#' @importFrom rvest html_nodes html_node html_text html_attr
#' @importFrom purrr map list_rbind
#' @importFrom tibble tibble
#' @importFrom stringr str_squish str_extract
wta_get_rankings <- function(type = c("singles", "doubles"), top = NULL,
                             wait = 12) {
  type <- match.arg(type)
  url  <- sprintf("https://www.wtatennis.com/rankings/%s", type)

  page <- .read_html_dynamic(url, wait = wait)
  out  <- .parse_rankings(page)

  if (!is.null(top)) out <- utils::head(out, top)
  out
}

#' @keywords internal
#' @noRd
.parse_rankings <- function(page) {
  rows <- rvest::html_nodes(page, "tr.player-row")
  if (length(rows) == 0L) {
    cli::cli_abort(c(
      "No ranking rows were found.",
      "i" = "WTA markup may have changed. Inspect {.code tr.player-row} in DevTools."
    ))
  }

  purrr::list_rbind(purrr::map(rows, function(r) {
    flag_cls <- rvest::html_attr(
      rvest::html_node(r, ".flag, .player-cell__flag"), "class"
    ) %||% ""
    cc <- stringr::str_extract(flag_cls, "flag--[A-Z]{3}")
    cc <- sub("flag--", "", cc %||% NA_character_)

    tibble::tibble(
      rank               = .text_or_na(r, ".player-row__cell--rank .player-row__rank"),
      player_id          = rvest::html_attr(r, "data-player-id"),
      player             = rvest::html_attr(r, "data-player-name"),
      country            = cc,
      age                = .text_or_na(r, ".player-row__cell--age"),
      tournaments_played = .text_or_na(r, ".player-row__cell--tournaments"),
      points             = .text_or_na(r, ".player-row__cell--points")
    )
  }))
}

read_ranking_fixture <- function() {
  xml2::read_html(testthat::test_path("fixtures", "rankings_singles.html"))
}

test_that(".parse_rankings extracts rows from the real tbody structure", {
  page <- read_ranking_fixture()
  out  <- matchpointR:::.parse_rankings(page)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 3L)
  expect_equal(out$rank,      c("1", "2", "3"))
  expect_equal(out$player_id, c("320760", "321091", "324166"))
  expect_equal(out$player,
               c("Aryna Sabalenka", "Iga Swiatek", "Elena Rybakina"))
  expect_equal(out$country,   c("BLR", "POL", "KAZ"))
  expect_equal(out$age,       c("27", "24", "26"))
  expect_equal(out$tournaments_played, c("20", "22", "18"))
  expect_equal(out$points,    c("11,025", "8,108", "6,450"))
})

test_that(".parse_rankings aborts on empty page", {
  page <- xml2::read_html("<html><body></body></html>")
  expect_error(matchpointR:::.parse_rankings(page), "No ranking rows")
})

# -------------------------------------------------------------------------
#  Player basic info -------------------------------------------------------

#' Get basic bio for a WTA player
#'
#' Parses the profile header of a WTA player page and returns a one-row
#' tibble with name, nationality, birth date, birth place, height and
#' handedness. The bulk of the data is read from the page's JSON-LD
#' (schema.org Person) block, which is more stable than the visual markup;
#' height is read from the profile bio block as a fallback.
#'
#' @param player_url Character. Full URL to a player page. Build it with
#'   [wta_player_url()] if you only have the numeric id.
#' @param download_images Logical. When `TRUE` (default) the headshot is
#'   downloaded into a `magick-image` object. Set to `FALSE` to skip the
#'   network round-trip and return only the image URL.
#'
#' @return A one-row [tibble::tibble()] with columns:
#' \describe{
#'   \item{`player_id`}{Numeric WTA id parsed from `@id`.}
#'   \item{`name`, `given_name`, `family_name`}{Name fields.}
#'   \item{`birth_date`}{Date of birth (ISO 8601 character).}
#'   \item{`nationality`, `birth_place`, `birth_country`}{Geography fields.}
#'   \item{`height`}{Height string as shown on the bio (e.g. `5' 9" (1.74m)`).}
#'   \item{`handedness`}{Dominant hand (`"Right-Handed"` / `"Left-Handed"`).}
#'   \item{`nationality_code`}{3-letter IOC/ISO code extracted from the flag
#'     image (e.g. `"CZE"`, `"USA"`).}
#'   \item{`player_image_url`, `nationality_flag_url`}{Headshot and flag URLs.}
#'   \item{`player_image`}{`magick-image` of the headshot, when
#'     `download_images = TRUE`.}
#'   \item{`nationality_flag`}{`magick-image` of the flag SVG, when
#'     `download_images = TRUE` and the suggested package \pkg{rsvg} is
#'     installed (otherwise `NA`).}
#' }
#' @export
#' @examplesIf interactive()
#' wta_get_player_basics(wta_player_url(320301, "katerina-siniakova"))
#' @importFrom rvest html_node html_nodes html_text html_attr
#' @importFrom tibble tibble
#' @importFrom stringr str_squish str_extract
wta_get_player_basics <- function(player_url, download_images = TRUE) {
  page <- .read_html_dynamic(player_url)
  .parse_player_basics(page, download_images = download_images)
}

#' Parse a WTA player profile from a pre-fetched page
#'
#' Internal parser so unit tests can hit local fixtures without launching
#' Chrome.
#'
#' @param page An `xml2::xml_document`.
#' @param download_images Logical, see [wta_get_player_basics()].
#' @keywords internal
#' @noRd
.parse_player_basics <- function(page, download_images = TRUE) {
  person <- .extract_jsonld(page, "Person")

  player_id <- NA_character_
  if (!is.null(person) && !is.null(person[["@id"]])) {
    player_id <- stringr::str_extract(person[["@id"]], "\\d+")
  }

  name        <- person[["name"]]        %||% NA_character_
  given_name  <- person[["givenName"]]   %||% NA_character_
  family_name <- person[["familyName"]]  %||% NA_character_
  birth_date  <- person[["birthDate"]]   %||% NA_character_
  nationality <- person[["nationality"]][["name"]]  %||% NA_character_
  addr        <- person[["birthPlace"]][["address"]]
  birth_place <- addr[["addressLocality"]] %||% NA_character_
  birth_ctry  <- addr[["addressCountry"]]  %||% NA_character_
  image_url   <- person[["image"]]       %||% NA_character_
  handedness  <- .prop_value(person[["additionalProperty"]], "Plays")

  height <- .bio_info(page, "Height")

  ## final safety net: fall back to HTML header if JSON-LD missing
  if (is.na(name)) {
    name <- .text_or_na(page, ".profile-header__name")
  }

  ## nationality flag (served as SVG from wtatennis.com)
  flag_url  <- .abs_url(.attr_or_na(page,
    ".profile-header__flag .flag__img, .profile-header-info__nationalityFlag",
    "src"))
  flag_code <- .attr_or_na(page,
    ".profile-header__flag .flag__img, .profile-header-info__nationalityFlag",
    "alt")

  out <- tibble::tibble(
    player_id            = player_id,
    name                 = as.character(name),
    given_name           = as.character(given_name),
    family_name          = as.character(family_name),
    birth_date           = as.character(birth_date),
    nationality          = as.character(nationality),
    nationality_code     = as.character(flag_code),
    birth_place          = as.character(birth_place),
    birth_country        = as.character(birth_ctry),
    height               = as.character(height),
    handedness           = as.character(handedness),
    player_image_url     = .abs_url(as.character(image_url)),
    nationality_flag_url = flag_url
  )

  if (download_images) {
    out$player_image     <- list(.safe_image_read(out$player_image_url))
    out$nationality_flag <- list(.safe_image_read_svg(out$nationality_flag_url))
  }
  out
}

#' @keywords internal
#' @noRd
.safe_image_read <- function(url) {
  if (is.na(url)) return(NA)
  tryCatch(magick::image_read(url), error = function(e) NA)
}

#' Read an SVG URL into a magick-image via \pkg{rsvg} (suggests)
#' @keywords internal
#' @noRd
.safe_image_read_svg <- function(url) {
  if (is.na(url)) return(NA)
  if (!requireNamespace("rsvg", quietly = TRUE)) return(NA)
  tryCatch(magick::image_read_svg(url), error = function(e) NA)
}

# -------------------------------------------------------------------------
#  Player overview (career highlights) -----------------------------------

#' Get a WTA player's career highlights
#'
#' Returns the structured "additional properties" block from the page's
#' JSON-LD: current singles and doubles rank, career titles, career prize
#' money. Supplements with the career-high singles rank read from the bio
#' side panel.
#'
#' @param player_url Character. URL to the player overview page.
#'
#' @return A long-format [tibble::tibble()] with columns `metric` and
#'   `value`. Rows include `singles_rank`, `doubles_rank`,
#'   `singles_career_titles`, `doubles_career_titles`,
#'   `career_prize_money`, `career_high`.
#' @export
#' @examplesIf interactive()
#' wta_get_player_overview(wta_player_url(320301, "katerina-siniakova"))
#' @importFrom tibble tibble
wta_get_player_overview <- function(player_url) {
  page <- .read_html_dynamic(player_url)
  .parse_player_overview(page)
}

#' @keywords internal
#' @noRd
.parse_player_overview <- function(page) {
  person <- .extract_jsonld(page, "Person")
  props  <- person[["additionalProperty"]]

  tibble::tibble(
    metric = c(
      "singles_rank", "doubles_rank",
      "singles_career_titles", "doubles_career_titles",
      "career_prize_money", "career_high"
    ),
    value = c(
      .prop_value(props, "WTA Singles Rank"),
      .prop_value(props, "WTA Doubles Rank"),
      .prop_value(props, "Singles Career Titles"),
      .prop_value(props, "Doubles Career Titles"),
      .prop_value(props, "Career Prize Money"),
      .bio_info(page, "Career High")
    )
  )
}

# -------------------------------------------------------------------------
#  Player matches ---------------------------------------------------------

#' Get the match history for a WTA player
#'
#' Walks the dynamic "Matches" page of a player profile, clicking the
#' "Show more" button until the full history is loaded, and returns one row
#' per match with tournament, round, opponent, score and result.
#'
#' @param player_url Character. URL to the player page; the function
#'   normalises to the `/matches` path automatically.
#' @param max_clicks Integer. Safety cap for the "Show more" click loop.
#'   Defaults to 50.
#'
#' @return A [tibble::tibble()] with one row per match and columns:
#'   `tournament`, `tournament_date`, `round`, `opponent`, `opponent_seed`,
#'   `opponent_country`, `opponent_rank`, `score`, `result`.
#' @export
#' @examplesIf interactive()
#' url <- wta_player_url(320301, "katerina-siniakova", "matches")
#' wta_get_player_matches(url)
#' @importFrom rvest html_nodes html_node html_text html_attr
#' @importFrom purrr map list_rbind
#' @importFrom tibble tibble
#' @importFrom stringr str_extract str_squish
wta_get_player_matches <- function(player_url, max_clicks = 50L) {
  player_url <- sub("#.*$", "", player_url)
  if (!grepl("/matches$", player_url)) {
    player_url <- paste0(sub("/$", "", player_url), "/matches")
  }

  page <- .read_html_dynamic(
    player_url,
    click_more_selector = ".player-matches__more-button",
    max_clicks = max_clicks
  )

  .parse_player_matches(page, source_url = player_url)
}

#' @keywords internal
#' @noRd
.parse_player_matches <- function(page, source_url = NA_character_) {
  tournaments <- rvest::html_nodes(page, ".player-matches__tournament")
  if (length(tournaments) == 0L) {
    cli::cli_abort(c(
      "No tournaments were found{ if (!is.na(source_url)) paste0(' at ', source_url) else '' }.",
      "i" = "WTA markup may have changed. Inspect {.code .player-matches__tournament} in DevTools."
    ))
  }

  rows <- purrr::list_rbind(purrr::map(tournaments, function(tour) {
    tournament      <- .text_or_na(tour, ".player-matches__tournament-title-link")
    tournament_date <- .attr_or_na(tour, ".player-matches__tournament-date time", "date-time")
    if (is.na(tournament_date)) {
      tournament_date <- .text_or_na(tour, ".player-matches__tournament-date")
    }

    matches <- rvest::html_nodes(
      tour,
      ".player-matches__match:not(.player-matches__match--divider)"
    )
    if (length(matches) == 0L) return(tibble::tibble())

    purrr::list_rbind(purrr::map(matches, function(m) {
      flag_node <- rvest::html_node(m, ".player-matches__match-opponent-flag")
      cc <- NA_character_
      if (!inherits(flag_node, "xml_missing") && !is.na(flag_node)) {
        cls <- rvest::html_attr(flag_node, "class") %||% ""
        cc <- stringr::str_extract(cls,
          "player-matches__match-opponent-flag--[A-Z]{3}")
        cc <- sub("player-matches__match-opponent-flag--", "", cc %||% NA)
      }

      result <- NA_character_
      if (length(rvest::html_nodes(m, ".player-matches__match-cell-result--win")) > 0L) {
        result <- "W"
      } else if (length(rvest::html_nodes(m, ".player-matches__match-cell-result--loss")) > 0L) {
        result <- "L"
      }

      tibble::tibble(
        tournament       = tournament,
        tournament_date  = tournament_date,
        round            = .text_or_na(m, ".player-matches__match-round"),
        opponent         = .text_or_na(m, ".player-matches__match-opponent-name"),
        opponent_seed    = .text_or_na(m, ".player-matches__match-opponent-seed"),
        opponent_country = cc,
        opponent_rank    = .text_or_na(m, ".player-matches__match-cell--opp-rank"),
        score            = .text_or_na(m, ".player-matches__match-cell--score"),
        result           = result
      )
    }))
  }))

  rows
}

read_fixture <- function(name) {
  xml2::read_html(testthat::test_path("fixtures", name))
}

test_that(".parse_player_basics reads JSON-LD + bio", {
  page <- read_fixture("player_profile.html")
  out  <- matchpointR:::.parse_player_basics(page, download_images = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)
  expect_equal(out$player_id,     "320301")
  expect_equal(out$name,          "Katerina Siniakova")
  expect_equal(out$given_name,    "Katerina")
  expect_equal(out$family_name,   "Siniakova")
  expect_equal(out$birth_date,    "1996-05-10")
  expect_equal(out$nationality,   "Czech Republic")
  expect_equal(out$birth_place,   "Hradec Kralove")
  expect_equal(out$birth_country, "Czech Republic")
  expect_equal(out$handedness,    "Right-Handed")
  expect_equal(out$height,        "5' 9\" (1.74m)")
  expect_equal(
    out$player_image_url,
    "https://wtafiles.blob.core.windows.net/images/headshots/320301.jpg"
  )
  expect_equal(out$nationality_code,     "CZE")
  expect_equal(
    out$nationality_flag_url,
    "https://www.wtatennis.com/resources/v7.32.16/i/elements/flags/cze.svg"
  )
})

test_that(".parse_player_overview builds the metric/value long table", {
  page <- read_fixture("player_profile.html")
  out  <- matchpointR:::.parse_player_overview(page)

  expect_s3_class(out, "tbl_df")
  expect_equal(ncol(out), 2L)
  expect_setequal(out$metric, c(
    "singles_rank", "doubles_rank",
    "singles_career_titles", "doubles_career_titles",
    "career_prize_money", "career_high"
  ))
  expect_equal(out$value[out$metric == "singles_rank"],          "42")
  expect_equal(out$value[out$metric == "doubles_rank"],          "2")
  expect_equal(out$value[out$metric == "singles_career_titles"], "5")
  expect_equal(out$value[out$metric == "doubles_career_titles"], "35")
  expect_equal(out$value[out$metric == "career_prize_money"],    "$15,787,557")
  expect_equal(out$value[out$metric == "career_high"],           "27")
})

test_that(".parse_player_matches extracts matches per tournament", {
  page <- read_fixture("player_matches.html")
  out  <- matchpointR:::.parse_player_matches(page)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 3L)
  expect_equal(out$tournament,
               c("Miami Open", "Miami Open", "BNP Paribas Open"))
  expect_equal(out$opponent,
               c("Jessica Pegula", "Paula Badosa", "Iga Swiatek"))
  expect_equal(out$opponent_country, c("USA", "ESP", "POL"))
  expect_equal(out$round,   c("F", "SF", "R16"))
  expect_equal(out$score,   c("6-3 6-4", "7-5 3-6 6-2", "4-6 3-6"))
  expect_equal(out$result,  c("W", "W", "L"))
  expect_equal(out$tournament_date[1], "2026-03-17")
})

test_that(".parse_player_matches aborts when no tournaments found", {
  page <- xml2::read_html("<html><body><p>nope</p></body></html>")
  expect_error(matchpointR:::.parse_player_matches(page),
               regexp = "No tournaments")
})

#' Build a WTA player URL
#'
#' Convenience wrapper to assemble a canonical player URL from a numeric id
#' and an optional slug.
#'
#' @param id Character or integer. The WTA numeric player id (e.g. `320301`).
#' @param slug Optional character. Player slug (e.g. `"katerina-siniakova"`).
#'   When omitted the URL still resolves — WTA redirects to the canonical one.
#' @param section Optional character. Page section to append as a path
#'   segment, one of `"overview"`, `"matches"`. Defaults to `"overview"`,
#'   which maps to the bare player URL.
#'
#' @return A single character string with the full URL.
#' @export
#' @examples
#' wta_player_url(320301, "katerina-siniakova")
#' wta_player_url(320301, "katerina-siniakova", "matches")
wta_player_url <- function(id, slug = NULL, section = c("overview", "matches")) {
  section <- match.arg(section)
  base <- sprintf("https://www.wtatennis.com/players/%s", id)
  if (!is.null(slug) && nzchar(slug)) base <- paste(base, slug, sep = "/")
  if (section == "overview") base else paste(base, section, sep = "/")
}

# -------------------------------------------------------------------------
#  Internal: dynamic HTML fetching ----------------------------------------

#' Fetch fully-rendered HTML with chromote
#'
#' Opens a headless Chrome session via \pkg{chromote}, waits for the page to
#' settle, optionally clicks a "load more" button and/or scrolls, and returns
#' the complete page source.
#'
#' @param url Character. Destination URL.
#' @param wait Numeric. Seconds to wait after initial navigation. Default 8.
#' @param click_more_selector Optional CSS selector for a "load more" button
#'   that should be clicked repeatedly until it disappears.
#' @param scroll Logical. Scroll to the bottom after each click? Default TRUE.
#' @param max_clicks Integer. Safety cap for the click loop. Default 50.
#' @param session Optional pre-existing `chromote::ChromoteSession`. When
#'   supplied it is reused (callers are responsible for closing it).
#'
#' @return A character string containing the full page source.
#' @keywords internal
#' @importFrom chromote ChromoteSession
.chromote_get_html <- function(url,
                               wait = 8,
                               click_more_selector = NULL,
                               scroll = TRUE,
                               max_clicks = 50L,
                               session = NULL) {
  ss <- session %||% chromote::ChromoteSession$new()
  if (is.null(session)) on.exit(try(ss$close(), silent = TRUE), add = TRUE)

  try(ss$Page$navigate(url, wait_ = FALSE), silent = TRUE)
  Sys.sleep(wait)

  if (scroll) {
    for (i in seq_len(3L)) {
      try(ss$Runtime$evaluate(
        "window.scrollTo(0, document.body.scrollHeight)"
      ), silent = TRUE)
      Sys.sleep(1)
    }
  }

  if (!is.null(click_more_selector)) {
    for (i in seq_len(max_clicks)) {
      present <- ss$Runtime$evaluate(
        sprintf("document.querySelector('%s') !== null", click_more_selector)
      )$result$value
      if (!isTRUE(present)) break
      ss$Runtime$evaluate(
        sprintf("document.querySelector('%s').click()", click_more_selector)
      )
      Sys.sleep(1.5)
      if (scroll) {
        ss$Runtime$evaluate("window.scrollTo(0, document.body.scrollHeight)")
        Sys.sleep(1)
      }
    }
  }

  ss$Runtime$evaluate("document.documentElement.outerHTML")$result$value
}

#' Read dynamic HTML into an xml2 document
#'
#' Thin wrapper around [`.chromote_get_html()`] that parses the rendered HTML
#' with [`xml2::read_html()`].
#'
#' @inheritParams .chromote_get_html
#' @return An `xml2::xml_document`.
#' @keywords internal
#' @importFrom xml2 read_html
.read_html_dynamic <- function(url, wait = 8, click_more_selector = NULL,
                               scroll = TRUE, max_clicks = 50L, session = NULL) {
  html <- .chromote_get_html(
    url = url, wait = wait,
    click_more_selector = click_more_selector,
    scroll = scroll, max_clicks = max_clicks, session = session
  )
  xml2::read_html(html)
}

# -------------------------------------------------------------------------
#  Internal: JSON-LD extraction -------------------------------------------

#' Extract a JSON-LD block of a given schema.org type
#'
#' The WTA site embeds several `<script type="application/ld+json">` blocks
#' per page. This helper parses each one and returns the first whose
#' `@type` equals `type`.
#'
#' @param page An `xml2::xml_document`.
#' @param type Character. schema.org type (e.g. `"Person"`).
#' @return A list parsed from JSON, or `NULL` if not found.
#' @keywords internal
#' @noRd
.extract_jsonld <- function(page, type) {
  nodes <- rvest::html_nodes(page, 'script[type="application/ld+json"]')
  for (n in nodes) {
    txt <- rvest::html_text(n)
    if (!nzchar(txt)) next
    parsed <- tryCatch(
      jsonlite::fromJSON(txt, simplifyVector = FALSE),
      error = function(e) NULL
    )
    if (is.null(parsed)) next
    t <- parsed[["@type"]]
    if (!is.null(t) && identical(t, type)) return(parsed)
  }
  NULL
}

# -------------------------------------------------------------------------
#  Internal: small helpers ------------------------------------------------

#' Null-coalescing operator
#' @keywords internal
#' @noRd
`%||%` <- function(a, b) if (is.null(a)) b else a

#' Safe text extraction from a single node
#' @keywords internal
#' @noRd
.text_or_na <- function(node, selector) {
  n <- rvest::html_node(node, selector)
  if (inherits(n, "xml_missing") || is.na(n)) return(NA_character_)
  stringr::str_squish(rvest::html_text(n, trim = TRUE))
}

#' Safe attribute extraction
#' @keywords internal
#' @noRd
.attr_or_na <- function(node, selector, attr) {
  n <- rvest::html_node(node, selector)
  if (inherits(n, "xml_missing") || is.na(n)) return(NA_character_)
  rvest::html_attr(n, attr)
}

#' Resolve a possibly-relative URL against the WTA base
#' @keywords internal
#' @noRd
.abs_url <- function(src, base = "https://www.wtatennis.com") {
  if (is.na(src) || !nzchar(src)) return(NA_character_)
  if (grepl("^https?://", src)) return(src)
  if (grepl("^//", src)) return(paste0("https:", src))
  paste0(base, src)
}

#' Pick the first non-empty value from `additionalProperty` by name
#'
#' @param props A list (as parsed from JSON-LD) of property-value objects.
#' @param nm Character. The `name` field to match.
#' @return Character scalar or `NA_character_`.
#' @keywords internal
#' @noRd
.prop_value <- function(props, nm) {
  if (is.null(props) || length(props) == 0L) return(NA_character_)
  for (p in props) {
    if (identical(p[["name"]], nm)) {
      v <- p[["value"]]
      if (!is.null(v) && length(v) == 1L && nzchar(as.character(v))) {
        return(as.character(v))
      }
    }
  }
  NA_character_
}

#' Extract the content of a `.profile-bio__info-block` by its title text
#'
#' @param page An `xml2::xml_document` or node.
#' @param title Character. Case-insensitive match against
#'   `.profile-bio__info-title`.
#' @return Character scalar or `NA_character_`.
#' @keywords internal
#' @noRd
.bio_info <- function(page, title) {
  blocks <- rvest::html_nodes(page, ".profile-bio__info-block")
  if (length(blocks) == 0L) return(NA_character_)
  titles <- vapply(blocks, .text_or_na, character(1),
                   selector = ".profile-bio__info-title")
  hit <- which(tolower(titles) == tolower(title))
  if (length(hit) == 0L) return(NA_character_)
  .text_or_na(blocks[[hit[1]]], ".profile-bio__info-content")
}

test_that("wta_player_url composes canonical URLs", {
  expect_equal(
    wta_player_url(320301, "katerina-siniakova"),
    "https://www.wtatennis.com/players/320301/katerina-siniakova"
  )
  expect_equal(
    wta_player_url(320301),
    "https://www.wtatennis.com/players/320301"
  )
  expect_equal(
    wta_player_url(320301, "katerina-siniakova", "matches"),
    "https://www.wtatennis.com/players/320301/katerina-siniakova/matches"
  )
  expect_error(wta_player_url(320301, section = "stats"))
})

test_that(".abs_url resolves relative, protocol-relative and absolute URLs", {
  expect_equal(
    matchpointR:::.abs_url("/flags/cz.svg"),
    "https://www.wtatennis.com/flags/cz.svg"
  )
  expect_equal(
    matchpointR:::.abs_url("//cdn.example.com/x.png"),
    "https://cdn.example.com/x.png"
  )
  expect_equal(
    matchpointR:::.abs_url("https://cdn.example.com/x.png"),
    "https://cdn.example.com/x.png"
  )
  expect_true(is.na(matchpointR:::.abs_url(NA_character_)))
  expect_true(is.na(matchpointR:::.abs_url("")))
})

test_that(".text_or_na / .attr_or_na return NA for missing nodes", {
  page <- xml2::read_html(
    "<html><body><p class='x'>hi</p><img class='y' src='/f.png'/></body></html>"
  )
  expect_equal(matchpointR:::.text_or_na(page, ".x"), "hi")
  expect_true(is.na(matchpointR:::.text_or_na(page, ".missing")))
  expect_equal(matchpointR:::.attr_or_na(page, ".y", "src"), "/f.png")
  expect_true(is.na(matchpointR:::.attr_or_na(page, ".missing", "src")))
})

test_that(".extract_jsonld finds the right schema.org block", {
  page <- xml2::read_html(paste0(
    '<html><head>',
    '<script type="application/ld+json">{"@type":"Organization","name":"X"}</script>',
    '<script type="application/ld+json">{"@type":"Person","name":"Y"}</script>',
    '</head><body></body></html>'
  ))
  person <- matchpointR:::.extract_jsonld(page, "Person")
  expect_equal(person$name, "Y")
  org <- matchpointR:::.extract_jsonld(page, "Organization")
  expect_equal(org$name, "X")
  expect_null(matchpointR:::.extract_jsonld(page, "Event"))
})

test_that(".prop_value picks the named value from additionalProperty", {
  props <- list(
    list(name = "Plays", value = "Right-Handed"),
    list(name = "WTA Singles Rank", value = "42")
  )
  expect_equal(matchpointR:::.prop_value(props, "Plays"), "Right-Handed")
  expect_equal(matchpointR:::.prop_value(props, "WTA Singles Rank"), "42")
  expect_true(is.na(matchpointR:::.prop_value(props, "Missing")))
  expect_true(is.na(matchpointR:::.prop_value(NULL, "Plays")))
})

test_that(".bio_info reads content by title", {
  page <- xml2::read_html(testthat::test_path("fixtures", "player_profile.html"))
  expect_equal(matchpointR:::.bio_info(page, "Height"), "5' 9\" (1.74m)")
  expect_equal(matchpointR:::.bio_info(page, "Career High"), "27")
  expect_true(is.na(matchpointR:::.bio_info(page, "Missing")))
})

#' matchpointR: Tidy Access to Women's Tennis Association (WTA) Data
#'
#' `matchpointR` is a small scraper toolkit that turns the public pages of
#' <https://www.wtatennis.com> into tidy data frames. It ships helpers for
#' player biographies, career highlights, full match histories and live
#' rankings.
#'
#' Dynamic content is rendered through a headless Chrome session using the
#' \pkg{chromote} package, so JavaScript-generated sections (matches,
#' rankings) are fully captured before parsing. Where possible the package
#' reads structured JSON-LD (schema.org) data instead of scraping CSS
#' classes, for resilience against site redesigns.
#'
#' @section Main functions:
#' * [wta_player_url()]
#' * [wta_get_player_basics()]
#' * [wta_get_player_overview()]
#' * [wta_get_player_matches()]
#' * [wta_get_rankings()]
#'
#' @section Author:
#' Alejandro Navas González (Angnar).
#'
#' @keywords internal
"_PACKAGE"

---
title: "Getting started with matchpointR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with matchpointR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(matchpointR)
```

## Overview

`matchpointR` turns the public pages of
[wtatennis.com](https://www.wtatennis.com) into tidy data frames:

| Function                      | Purpose                                                   |
|-------------------------------|-----------------------------------------------------------|
| `wta_player_url()`            | Build canonical player URLs.                              |
| `wta_get_player_basics()`     | One-row tibble with bio parsed from the page's JSON-LD.   |
| `wta_get_player_overview()`   | Career highlights (ranks, titles, prize money).           |
| `wta_get_player_matches()`    | One row per match across the full career.                 |
| `wta_get_rankings()`          | Current singles / doubles leaderboard.                    |

Every function that hits the network opens (and closes) its own headless
Chrome session through `chromote`. Where the WTA site exposes structured
schema.org JSON-LD data, `matchpointR` reads from that in preference to
CSS selectors — this is substantially more resilient against site
redesigns.

## A worked example: Kateřina Siniaková

```{r}
url <- wta_player_url(320301, "katerina-siniakova")

bio <- wta_get_player_basics(url, download_images = FALSE)
bio
```

`download_images = TRUE` (the default) additionally downloads the
headshot into a `magick-image` list-column — set it to `FALSE` when you
only need the metadata.

### Career highlights

```{r}
wta_get_player_overview(url)
```

Returns a long tibble with one row per metric:
`singles_rank`, `doubles_rank`, `singles_career_titles`,
`doubles_career_titles`, `career_prize_money`, `career_high`.

### Full match history

```{r}
matches_url <- wta_player_url(320301, "katerina-siniakova", "matches")
matches <- wta_get_player_matches(matches_url)
head(matches)
```

`wta_get_player_matches()` clicks the *Show more* button repeatedly
until no more matches are loaded. Raise `max_clicks` only if you hit
the safety cap on a very long career.

## Tour-wide data

### Live rankings

```{r}
wta_get_rankings("singles", top = 50)
wta_get_rankings("doubles", top = 50)
```

## Tips

- **Coercion.** All columns are returned as character to stay faithful
  to the rendered page. Cast to numeric/date downstream.
- **Selector drift.** If a function suddenly returns zero rows, the
  site's CSS classes have probably been renamed — open an issue.
- **Rate limiting.** `chromote` hits the site as a real browser; be
  considerate if you are iterating over many players.

## Not yet covered

Tour-wide statistics leaderboards (aces, winners, break-points
converted, …) are tracked in
[issue #1](https://github.com/Angnar-97/matchpointR/issues/1) for a
future release. The WTA site reshuffled its `/stats` hub and we are
waiting for a stable URL pattern before committing to an API.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matchpointR-package.R
\docType{package}
\name{matchpointR-package}
\alias{matchpointR}
\alias{matchpointR-package}
\title{matchpointR: Tidy Access to Women's Tennis Association (WTA) Data}
\description{
\code{matchpointR} is a small scraper toolkit that turns the public pages of
\url{https://www.wtatennis.com} into tidy data frames. It ships helpers for
player biographies, career highlights, full match histories and live
rankings.
}
\details{
Dynamic content is rendered through a headless Chrome session using the
\pkg{chromote} package, so JavaScript-generated sections (matches,
rankings) are fully captured before parsing. Where possible the package
reads structured JSON-LD (schema.org) data instead of scraping CSS
classes, for resilience against site redesigns.
}
\section{Main functions}{

\itemize{
\item \code{\link[=wta_player_url]{wta_player_url()}}
\item \code{\link[=wta_get_player_basics]{wta_get_player_basics()}}
\item \code{\link[=wta_get_player_overview]{wta_get_player_overview()}}
\item \code{\link[=wta_get_player_matches]{wta_get_player_matches()}}
\item \code{\link[=wta_get_rankings]{wta_get_rankings()}}
}
}

\section{Author}{

Alejandro Navas González (Angnar).
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Angnar-97/matchpointR}
  \item Report bugs at \url{https://github.com/Angnar-97/matchpointR/issues}
}

}
\author{
\strong{Maintainer}: Alejandro Navas González \email{angnar@telaris.es} (Angnar)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/players.R
\name{wta_get_player_overview}
\alias{wta_get_player_overview}
\title{Get a WTA player's career highlights}
\usage{
wta_get_player_overview(player_url)
}
\arguments{
\item{player_url}{Character. URL to the player overview page.}
}
\value{
A long-format \code{\link[tibble:tibble]{tibble::tibble()}} with columns \code{metric} and
\code{value}. Rows include \code{singles_rank}, \code{doubles_rank},
\code{singles_career_titles}, \code{doubles_career_titles},
\code{career_prize_money}, \code{career_high}.
}
\description{
Returns the structured "additional properties" block from the page's
JSON-LD: current singles and doubles rank, career titles, career prize
money. Supplements with the career-high singles rank read from the bio
side panel.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
wta_get_player_overview(wta_player_url(320301, "katerina-siniakova"))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.chromote_get_html}
\alias{.chromote_get_html}
\title{Fetch fully-rendered HTML with chromote}
\usage{
.chromote_get_html(
  url,
  wait = 8,
  click_more_selector = NULL,
  scroll = TRUE,
  max_clicks = 50L,
  session = NULL
)
}
\arguments{
\item{url}{Character. Destination URL.}

\item{wait}{Numeric. Seconds to wait after initial navigation. Default 8.}

\item{click_more_selector}{Optional CSS selector for a "load more" button
that should be clicked repeatedly until it disappears.}

\item{scroll}{Logical. Scroll to the bottom after each click? Default TRUE.}

\item{max_clicks}{Integer. Safety cap for the click loop. Default 50.}

\item{session}{Optional pre-existing \code{chromote::ChromoteSession}. When
supplied it is reused (callers are responsible for closing it).}
}
\value{
A character string containing the full page source.
}
\description{
Opens a headless Chrome session via \pkg{chromote}, waits for the page to
settle, optionally clicks a "load more" button and/or scrolls, and returns
the complete page source.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/players.R
\name{wta_get_player_matches}
\alias{wta_get_player_matches}
\title{Get the match history for a WTA player}
\usage{
wta_get_player_matches(player_url, max_clicks = 50L)
}
\arguments{
\item{player_url}{Character. URL to the player page; the function
normalises to the \verb{/matches} path automatically.}

\item{max_clicks}{Integer. Safety cap for the "Show more" click loop.
Defaults to 50.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} with one row per match and columns:
\code{tournament}, \code{tournament_date}, \code{round}, \code{opponent}, \code{opponent_seed},
\code{opponent_country}, \code{opponent_rank}, \code{score}, \code{result}.
}
\description{
Walks the dynamic "Matches" page of a player profile, clicking the
"Show more" button until the full history is loaded, and returns one row
per match with tournament, round, opponent, score and result.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
url <- wta_player_url(320301, "katerina-siniakova", "matches")
wta_get_player_matches(url)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rankings.R
\name{wta_get_rankings}
\alias{wta_get_rankings}
\title{Get the current WTA rankings}
\usage{
wta_get_rankings(type = c("singles", "doubles"), top = NULL, wait = 12)
}
\arguments{
\item{type}{Character. One of \code{"singles"}, \code{"doubles"}. Defaults to
\code{"singles"}.}

\item{top}{Integer. Limit the output to the top \code{N} ranked players.
\code{NULL} (default) keeps every row rendered by the page.}

\item{wait}{Numeric. Seconds to wait for the rankings widget to hydrate
after navigation. Defaults to 12.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} with one row per player and columns:
\code{rank}, \code{player_id}, \code{player}, \code{country}, \code{age},
\code{tournaments_played}, \code{points}.
}
\description{
Scrapes the rankings table at
\url{https://www.wtatennis.com/rankings/singles} (or \verb{/doubles}) and returns
a tidy tibble. The initial page renders the first 50 rows; increase the
browser dwell time with \code{wait} if the widget hasn't hydrated yet.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
wta_get_rankings("singles", top = 50)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/players.R
\name{wta_get_player_basics}
\alias{wta_get_player_basics}
\title{Get basic bio for a WTA player}
\usage{
wta_get_player_basics(player_url, download_images = TRUE)
}
\arguments{
\item{player_url}{Character. Full URL to a player page. Build it with
\code{\link[=wta_player_url]{wta_player_url()}} if you only have the numeric id.}

\item{download_images}{Logical. When \code{TRUE} (default) the headshot is
downloaded into a \code{magick-image} object. Set to \code{FALSE} to skip the
network round-trip and return only the image URL.}
}
\value{
A one-row \code{\link[tibble:tibble]{tibble::tibble()}} with columns:
\describe{
\item{\code{player_id}}{Numeric WTA id parsed from \verb{@id}.}
\item{\code{name}, \code{given_name}, \code{family_name}}{Name fields.}
\item{\code{birth_date}}{Date of birth (ISO 8601 character).}
\item{\code{nationality}, \code{birth_place}, \code{birth_country}}{Geography fields.}
\item{\code{height}}{Height string as shown on the bio (e.g. \verb{5' 9" (1.74m)}).}
\item{\code{handedness}}{Dominant hand (\code{"Right-Handed"} / \code{"Left-Handed"}).}
\item{\code{nationality_code}}{3-letter IOC/ISO code extracted from the flag
image (e.g. \code{"CZE"}, \code{"USA"}).}
\item{\code{player_image_url}, \code{nationality_flag_url}}{Headshot and flag URLs.}
\item{\code{player_image}}{\code{magick-image} of the headshot, when
\code{download_images = TRUE}.}
\item{\code{nationality_flag}}{\code{magick-image} of the flag SVG, when
\code{download_images = TRUE} and the suggested package \pkg{rsvg} is
installed (otherwise \code{NA}).}
}
}
\description{
Parses the profile header of a WTA player page and returns a one-row
tibble with name, nationality, birth date, birth place, height and
handedness. The bulk of the data is read from the page's JSON-LD
(schema.org Person) block, which is more stable than the visual markup;
height is read from the profile bio block as a fallback.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
wta_get_player_basics(wta_player_url(320301, "katerina-siniakova"))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.read_html_dynamic}
\alias{.read_html_dynamic}
\title{Read dynamic HTML into an xml2 document}
\usage{
.read_html_dynamic(
  url,
  wait = 8,
  click_more_selector = NULL,
  scroll = TRUE,
  max_clicks = 50L,
  session = NULL
)
}
\arguments{
\item{url}{Character. Destination URL.}

\item{wait}{Numeric. Seconds to wait after initial navigation. Default 8.}

\item{click_more_selector}{Optional CSS selector for a "load more" button
that should be clicked repeatedly until it disappears.}

\item{scroll}{Logical. Scroll to the bottom after each click? Default TRUE.}

\item{max_clicks}{Integer. Safety cap for the click loop. Default 50.}

\item{session}{Optional pre-existing \code{chromote::ChromoteSession}. When
supplied it is reused (callers are responsible for closing it).}
}
\value{
An \code{xml2::xml_document}.
}
\description{
Thin wrapper around \code{\link[=.chromote_get_html]{.chromote_get_html()}} that parses the rendered HTML
with \code{\link[xml2:read_xml]{xml2::read_html()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{wta_player_url}
\alias{wta_player_url}
\title{Build a WTA player URL}
\usage{
wta_player_url(id, slug = NULL, section = c("overview", "matches"))
}
\arguments{
\item{id}{Character or integer. The WTA numeric player id (e.g. \code{320301}).}

\item{slug}{Optional character. Player slug (e.g. \code{"katerina-siniakova"}).
When omitted the URL still resolves — WTA redirects to the canonical one.}

\item{section}{Optional character. Page section to append as a path
segment, one of \code{"overview"}, \code{"matches"}. Defaults to \code{"overview"},
which maps to the bare player URL.}
}
\value{
A single character string with the full URL.
}
\description{
Convenience wrapper to assemble a canonical player URL from a numeric id
and an optional slug.
}
\examples{
wta_player_url(320301, "katerina-siniakova")
wta_player_url(320301, "katerina-siniakova", "matches")
}
